(*
 Oukseh Lee, The LET Project, KAIST

 This source is almost copied from the Objective Caml compiler.
*)



(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)


signature NTCOMPILE = sig
  val implementation: Format.formatter -> string -> unit
  val c_file: string -> unit
  val initial_env: unit -> Env.t
  val init_path: unit -> unit
  val compile_signature: Format.formatter -> string -> unit
  val create_archive: string list -> string -> unit
end

structure NTcompile : NTCOMPILE = struct

open Misc
open Config
open Format
open Typedtree

(* Initialize the search path.
   The current directory is always searched first,
   then the directories specified with the -I option (in command-line order),
   then the standard library directory. *)

fun init_path () =
  let val dirs =
        if !Clflags.thread_safe then
           "+threads" :: !Clflags.include_dirs
        else
           !Clflags.include_dirs 
      val exp_dirs =
        List.map (expand_directory Config.standard_library) dirs 
  in  load_path := ("" :: List.rev (Config.standard_library :: exp_dirs));
      Env.reset_cache()
  end

(* Return the initial environment in which compilation proceeds. *)

fun initial_env () = (
    (**** cookcu ****)
    if !Clflags.ntyping then Init.init (!Clflags.include_dirs @ ["./"]);
    (**** cookcu ****)
    if !Clflags.nopervasives
    then Env.initial
    else if !Clflags.ntyping then
             Env.open_pers_signature "_n_typing" 
	     (Env.open_pers_signature "_n_setmaphash"
             (Env.open_pers_signature "Pervasives" Env.initial))
    else Env.open_pers_signature "_n_setmaphash"
             (Env.open_pers_signature "Pervasives" Env.initial))
  handle Not_found =>
    fatal_error "cannot open pervasives.cmi"

(* Optionally preprocess a source file *)

fun preprocess sourcefile tmpfile =
  case !Clflags.preprocessor of
    None => sourcefile
  | Some pp =>
      let val comm = pp ^ " " ^ sourcefile ^ " > " ^ tmpfile 
      in  if Ccomp.command comm <> 0 then (
            Printf.eprintf "Preprocessing error\n";
            flush stderr;
            exit 2);
          tmpfile
      end

fun remove_preprocessed inputfile =
  case !Clflags.preprocessor of
    None => ()
  | Some _ => remove_file inputfile

(* Parse a file or get a dumped syntax tree in it *)

exception Outdated_version

fun parse_file inputfile parse_fun ast_magic =
  let val ic = open_in_bin inputfile
      val is_ast_file =
        let val buffer = String.create (String.length ast_magic)
        in  really_input ic buffer 0 (String.length ast_magic);
            if buffer = ast_magic then true
            else if String.sub buffer 0 9 = String.sub ast_magic 0 9 then 
              raise Outdated_version
            else false
        end handle Outdated_version =>
                     fatal_error "Ocaml and preprocessor have incompatible versions"
                 | _ => false  
      val ast = (if is_ast_file then (
                  Location.input_name := input_value ic;
                  input_value ic)
                else (
                  seek_in ic 0;
                  Location.input_name := inputfile;
                  parse_fun (Lexing.from_channel ic)))
                handle x => (close_in ic; raise x)
  in  close_in ic;
      ast
  end
(*
fun typing sourcefile prefixname modulename env ast =
  let val ntyfile = prefixname ^ ".nty"
  in  ignore(Infer.infer ast);
      Dump.dump ntyfile;
      Ntypemod.type_implementation sourcefile prefixname modulename env ast
  end     
*)
fun print_if ppf flag printer arg = 
  if !flag then fprintf ppf "%a@." printer arg
  
fun implementation ppf sourcefile = let
  val _ = init_path()
  val prefixname = Filename.chop_extension sourcefile 
  val modulename = (Filename.basename prefixname) 
  val inputfile = preprocess sourcefile (prefixname ^ ".ppo") 
  val objfile = prefixname ^ ".cmo" 
  val ntyfile = prefixname ^ ".nty" 
  val oc = open_out_bin objfile 
  val env = initial_env() 
  (***** poisson start *****)
  val _ = Strtbl_build.build_str_table inputfile (!load_path)
  fun implementation_parse lb =
    let val nast = Nparse.implementation_parse lb
        val nast' = if !Clflags.ntyping then
                      let val _ = NEnv.reset_file modulename
                          val absyn = Infer.infer nast
                      in  ignore(NEnv.stabilize_file());
                          (* EnvPrint.b !(NEnv.s2) EnvPrint.none; *)
                          Dump.dump ntyfile;
                          Absyn2Ast.topdec (Remove_record.topdec absyn)
                      end
                    else nast
    in  N2caml.trans_topdec (Nconvert.convert_topdec nast')
    end
  (***** poisson end *******)
  in let val i0 = parse_file inputfile implementation_parse ast_impl_magic_number
         val _  = print_if ppf Clflags.dump_parsetree Printast.implementation i0
         val i1 = Typemod.type_implementation sourcefile prefixname modulename env i0
         val i2 = Translmod.transl_implementation modulename i1
         val _  = print_if ppf Clflags.dump_rawlambda Printlambda.lambda i2
         val i3 = Simplif.simplify_lambda i2
         val _  = print_if ppf Clflags.dump_lambda Printlambda.lambda i3
         val i4 = Bytegen.compile_implementation modulename i3
     in  print_if ppf Clflags.dump_instr Printinstr.instrlist i4;
         Emitcode.to_file oc modulename i4;
         remove_preprocessed inputfile;
         close_out oc
     end handle x => (close_out oc; remove_file objfile; remove_file ntyfile; raise x)
  end

fun compile_signature ppf sourcefile = let
  val _ = (init_path(); NEnv.sig_context_prefix := "")
  val prefixname = Filename.chop_extension sourcefile 
  val ntyfile = prefixname ^ ".nty"
  val env = initial_env()
  fun implementation_parse lb =
    let val nast = Nparse.implementation_parse lb
    in  if !Clflags.ntyping then (
          ignore(Infer.infer nast);
          NEnv.convert_to_str();
          (* EnvPrint.b (!(NEnv.s2)) EnvPrint.none; *)
          Dump.dump ntyfile)
    end
  in  parse_file sourcefile implementation_parse ast_impl_magic_number
  end

fun create_archive files outfile = let
 val bs = List.iter (fn f => let val (b,c) = Dump.load f
                             in  NEnv.add_b b;
                                 Ty.CE.add_ce' c
                             end) files
 in  Dump.dump outfile
 end

fun c_file name = if Ccomp.compile_file name <> 0 then exit 2

end  

